#include <string.h>
#include <stdio.h>
#include "project.h"
#include "system.h"

#define CONST_LCD_NUM_ROWS 4
#define CONST_LCD_WIDTH 20
#define CONST_MAX_LCD_CONTRAST 160
#define CONST_MAX_LCD_BACKLIGHT 160



/////////////////////////////////////////////////
//Communication modes
/////////////////////////////////////////////////
#define CONST_MODE_I2C              0
#define CONST_MODE_UART_9600        1
#define CONST_MODE_UART_19200       2
#define CONST_MODE_UART_38400       3
#define CONST_MODE_UART_57600       4
#define CONST_MODE_UART_115200      5
#define CONST_MODE_SPI_CPOL0_CPHA0    6
#define CONST_MODE_SPI_CPOL0_CPHA1    7
#define CONST_MODE_SPI_CPOL1_CPHA0    8
#define CONST_MODE_SPI_CPOL1_CPHA1    9

/////////////////////////////////////////////////
//Unlock commands
/////////////////////////////////////////////////
#define CONST_SPL_FUNC_CPU_RESET                0
#define CONST_SPL_FUNC_LCD_CONTRAST         		3
#define CONST_SPL_FUNC_LCD_CONTRAST_SAVE        4
#define CONST_SPL_FUNC_LCD_BACKLIGHT        		5
#define CONST_SPL_FUNC_LCD_BACKLIGHT_SAVE       6
#define CONST_SPL_FUNC_I2C                      13
#define CONST_SPL_FUNC_INTERFACE                14



/////////////////////////////////////////////////
//Registers
/////////////////////////////////////////////////
#define CONST_LCD_DATA_SIZE             21
#define CONST_LCD_ACTION_GO_BIT 			(1 << 7)
#define CONST_ERR_FLASH_BIT            (1 << 0)
#define CONST_LCD_EMPTY 0x20
#define CONST_UNLOCK1 0x12
#define CONST_UNLOCK2 0x34 
#define CONST_ADDR_ID_REG_DEF           0x3F 



#define CONST_ADDR_ID_REV               0
#define CONST_ADDR_BTL_REV              1
#define CONST_ADDR_PCB_REV              2
#define CONST_ADDR_PCBA_REV             3
#define CONST_ADDR_FW_REV               4
#define CONST_ADDR_I2C_DEV_ADDR         5
#define CONST_ADDR_LCD_CONTRAST         6
#define CONST_ADDR_LCD_BACKLIGHT        7
#define CONST_ADDR_LCD_NUM_ROWS         8
#define CONST_ADDR_LCD_CHARS_PER_ROW    9
#define CONST_ADDR_LCD_R0_ADDR          10
#define CONST_ADDR_LCD_R1_ADDR          11
#define CONST_ADDR_LCD_R2_ADDR          12
#define CONST_ADDR_LCD_R3_ADDR          13
#define CONST_ADDR_INTR_ADDR            14
#define CONST_ADDR_ERR_REG              15

//Above here are read only registers
#define CONST_ADDR_I2C_READ_REG         16
#define CONST_ADDR_SPCL_W_REG_1         17
#define CONST_ADDR_SPCL_REG             18
#define CONST_ADDR_UNLOCK1_REG          19
#define CONST_ADDR_UNLOCK2_REG          20

#define CONST_ADDR_LCD_ACTION           21
#define CONST_ADDR_LCD_DATA_START       22
#define CONST_ADDR_LCD_DATA_END         (CONST_ADDR_LCD_DATA_START + CONST_LCD_DATA_SIZE)


#define CONST_LAST_REG_ADDR     CONST_ADDR_LCD_DATA_END
#define CONST_FIRST_WRITE_REG_ADDR  CONST_ADDR_I2C_READ_REG

//Wait 5mS for LCD to update
#define LCD_WAIT_ROW_UPDATE   {sys_delay_ms(5);}
	
//Delay 5mS after special unlock sequence commands
#define LCD_WAIT_SPL_FUNC  {sys_delay_ms(5);}

//Wait 200uS at the end of every packet
#define LCD_END_PACKET_DELAY {sys_delay_ms(1);}

//Wait 100uS
#define LCD_UART_RX_DELAY {sys_delay_ms(1);}

//Wait 100uS
#define LCD_UART_TX_DELAY {sys_delay_ms(1);}


//Wait 200uS
#define LCD_SPI_TX_DELAY  {sys_delay_ms(1);}

#define CONST_LCD_UART_REG_READ (0x00)
#define CONST_LCD_UART_REG_WRITE (0x80)    



/////////////////////////////////////////////////
//Prototypes
/////////////////////////////////////////////////
U8 isolate_reg_bits(U8 val, U8 bits);
U8 lcdSpecialFunction(U8 func);
U8 lcdBurstReadReg(U8 start_addr, U8 len, U8 *data);
U8 lcdReadReg(U8 addr, U8 *data);
U8 lcdWriteReg(U8 addr, U8 data);
U8 lcdRowWrite(U8 row, U8 *data);
U8 lcdReset(void);
U8 lcdContrast(U8 val);
U8 lcdContrastSave(void);
U8 lcdBacklight(U8 val);
U8 lcdBacklightSave(void);
U8 lcdChangeI2cSlaveAddr(U8 addr); 
U8 init_coms(void);
U8 lcd_i2c_write_reg(U8 reg_addr, U8 data);
U8 lcd_i2c_burst_read_reg(U8 start_addr, U8 len, U8 *data);
U8 lcd_i2c_setup_reg_addr(U8 addr);
U8 lcd_uart_burst_read_reg(U8 start_addr, U8 len, U8 *data);
U8 lcd_uart_tx(U8 *buf, U8 num_bytes);
U8 lcd_uart_write_reg(U8 addr, U8 data);
U8 lcd_spi_write_reg(U8 addr, U8 data);
U8 lcd_spi_burst_read_reg(U8 start_addr, U8 len, U8 *data);


/////////////////////////////////////////////////
//Operating System Specific Functions
/////////////////////////////////////////////////


#if defined(I2C_MODE)


U8 lcd_i2c_write_reg(U8 reg_addr, U8 data)
{	U8 x;
	U8 buf[2];
		
	buf[0] = reg_addr;
	buf[1] = data;
		
	x = sys_i2c_write(CONST_I2C_SLAVE_ADDR, buf, 2);
	LCD_END_PACKET_DELAY;
	return x;
}



U8 lcd_i2c_burst_read_reg(U8 start_addr, U8 len, U8 *data)
{ 	U8 x;
		
	x = lcd_i2c_setup_reg_addr(start_addr);
	if(x != CONST_SUCCESS)
	{	return x;
	}
	
	x = sys_i2c_read(CONST_I2C_SLAVE_ADDR, data, len);
	LCD_END_PACKET_DELAY;
	return x;
}

U8 lcd_i2c_setup_reg_addr(U8 addr)
{	U8 x;
	U8 buf[2];
		
	buf[0] = (U8)CONST_ADDR_I2C_READ_REG;
	buf[1] = (U8)addr;
	 
	//write addr to offset CONST_ADDR_I2C_ID_REG to setup address
	x = sys_i2c_write(CONST_I2C_SLAVE_ADDR, buf, 2);
	LCD_END_PACKET_DELAY;
	return x;
}


#endif


#if defined(UART_MODE)

U8 lcd_uart_burst_read_reg(U8 start_addr, U8 len, U8 *data)
{	U8 i;
	U8 tmp;

	for(i = 0; i < len; i++)
	{	//Send read request
		tmp = CONST_LCD_UART_REG_READ | (start_addr+i);
		if(sys_uart_tx_byte(tmp) != CONST_SUCCESS)
		{	return CONST_ERR;
		}
		
		//Wait for reply
		if(sys_uart_rx_byte(&tmp) != CONST_SUCCESS)
		{	return CONST_ERR;
        }
        
        LCD_UART_RX_DELAY;
		
		//Store the value
		data[i] = tmp;
	}
	
	LCD_END_PACKET_DELAY;
	return CONST_SUCCESS;
} 


U8 lcd_uart_tx(U8 *buf, U8 num_bytes)
{	U8 x, y;

	y = 0;
	
	for(x = 0; x < num_bytes; x++)
	{	y |= sys_uart_tx_byte(buf[x]);
		LCD_UART_TX_DELAY;
	}
	
	return y;
}
 


U8 lcd_uart_write_reg(U8 addr, U8 data)
{ 	U8 x;
	U8 buf[2];

	//Send read request
	buf[0] = CONST_LCD_UART_REG_WRITE | addr;
	buf[1] = data;
	
	x  = lcd_uart_tx(buf, 2);
	LCD_END_PACKET_DELAY;
	
	return x;
}

#endif


#if defined(SPI_MODE)

U8 lcd_spi_write_reg(U8 addr, U8 data)
{	U8 x;
	U8 rx;

	sys_spi_cs(0);
    
    //Send address
	x = sys_spi_byte(CONST_LCD_UART_REG_WRITE | addr, &rx);
    LCD_SPI_TX_DELAY;
   
  	//Send data;
	x |= sys_spi_byte(data, &rx);
	sys_spi_cs(1);
    
    LCD_SPI_TX_DELAY;
 	LCD_END_PACKET_DELAY;
	return x;
}
 
U8 lcd_spi_burst_read_reg(U8 start_addr, U8 len, U8 *data)
{	U8 x, y;
  	U8 rx;
  	
  	y = 0;
	  
	sys_spi_cs(0);
    for(x = 0; x < len; x++)
	{	//Send read request
		y |= sys_spi_byte(CONST_LCD_UART_REG_READ | (start_addr+x), &rx);
        LCD_SPI_TX_DELAY;
  	   
	  	//Read data
		y |= sys_spi_byte(0, &rx);
  		data[x] = rx;
  		LCD_SPI_TX_DELAY;
  		LCD_END_PACKET_DELAY;
	}	
    sys_spi_cs(1);
  
  
  	return y;
}

#endif





U8 isolate_reg_bits(U8 val, U8 bits)
{ return val & bits;
}  


//Special operations requiring an unlock sequence
U8 lcdSpecialFunction(U8 func)
{	U8 x;

	x = lcdWriteReg(CONST_ADDR_SPCL_REG, func);	
	x |= lcdWriteReg(CONST_ADDR_UNLOCK1_REG, CONST_UNLOCK1);
	x |= lcdWriteReg(CONST_ADDR_UNLOCK2_REG, CONST_UNLOCK2);	
  
  	//Wait a little bit 
  	LCD_WAIT_SPL_FUNC;
	return x;
}

//Burst Read register
U8 lcdBurstReadReg(U8 start_addr, U8 len, U8 *data)
{  U8 x = CONST_ERR;
	
	#if defined(I2C_MODE)
	x = lcd_i2c_burst_read_reg(start_addr, len, data);
		
	#elif defined(UART_MODE)
	x = lcd_uart_burst_read_reg(start_addr, len, data);
	
	#else//SPI_MODE
	x = lcd_spi_burst_read_reg(start_addr, len, data);
	#endif
	
	return x;
}


//Read register
U8 lcdReadReg(U8 addr, U8 *data)
{ 	return lcdBurstReadReg(addr,1,data);
}


//Write Register
U8 lcdWriteReg(U8 addr, U8 data)
{  U8 x = CONST_ERR;
	
	#if defined(I2C_MODE)
	x = lcd_i2c_write_reg(addr, data);
		
	#elif defined(UART_MODE)
	x = lcd_uart_write_reg(addr, data);
	
	#else//SPI_MODE
	x = lcd_spi_write_reg(addr,data);
	
	#endif
	
	return x;
}





//Update the LCD
//row should be 1-4
//Data is a pointer to the data to be displayed on that particular row
//
//To clear the LCD, update each row with white space " "
U8 lcdRowWrite(U8 row, U8 *new_data)
{	U8 x,y;
	U8 data_string[CONST_LCD_WIDTH+1];
    
  	//Check row number
	if((row < 1) || (row > CONST_LCD_NUM_ROWS))
  	{ 	return CONST_ERR;
	}
	  
	//Check length of string
	y = strlen(new_data);
  	if(y > CONST_LCD_WIDTH)
 	{ 	return CONST_ERR;
	}
	
	//Initialize to empty
  	memset(data_string, CONST_LCD_EMPTY, sizeof(data_string));
   
  	//copy the new data
  	memcpy(data_string, new_data, y);
 
  	//Write to LCD
  	x = 0;
	for(y = 0; y < CONST_LCD_WIDTH; y++)
  	{ x |= lcdWriteReg(CONST_ADDR_LCD_DATA_START + y, data_string[y]);
  	}
 
  	//Issue update command
  	x |= lcdWriteReg(CONST_ADDR_LCD_ACTION, (row-1) | CONST_LCD_ACTION_GO_BIT);

   //Wait for LCD update
	LCD_WAIT_ROW_UPDATE;
     
  	return x;
}

//Reset the LCD
U8 lcdReset(void)
{	return lcdSpecialFunction(CONST_SPL_FUNC_CPU_RESET); 
}

//Updates the contrast, argument can be between 0 and 160
//Changes are not saved to non volatile memory
U8 lcdContrast(U8 val)
{	U8 y;

	y = lcdWriteReg(CONST_ADDR_SPCL_W_REG_1, val);
	y |= lcdSpecialFunction(CONST_SPL_FUNC_LCD_CONTRAST); 
	return y;
}

//Save the contrast setting to non volatile memory
//
//Note:
//In applications where the contrast is frequently adjusted, please use
//the lcdContrast function alone and don't save the settings after each change.
//Saving the settings very freqeuently can prematurely wear out the non volatile memory.
U8 lcdContrastSave(void)
{	return lcdSpecialFunction(CONST_SPL_FUNC_LCD_CONTRAST_SAVE);
}

//Update the backlight, argument can be between 0 and 160
//Changes are not saved to non volatile memory
U8 lcdBacklight(U8 val)
{	U8 y;

	y = lcdWriteReg(CONST_ADDR_SPCL_W_REG_1, val);
  	y |= lcdSpecialFunction(CONST_SPL_FUNC_LCD_BACKLIGHT); 
  	return y;
}

//Save the backlight setting to non volatile memory.
//
//Note:
//In applications where the backlight is frequently adjusted, please use
//the lcdBacklight function alone and don't save the settings after each change.
//Saving the settings very freqeuently can prematurely wear out the non volatile memory.
U8 lcdBacklightSave(void)
{	return lcdSpecialFunction(CONST_SPL_FUNC_LCD_BACKLIGHT_SAVE);
}

//Changes the i2c slave address, saves changes to non volatile memory and
//resets the LCD for the changes to take effect.
U8 lcdChangeI2cSlaveAddr(U8 addr)
{	U8 y;

	if(addr > 0x7F)
	{	return CONST_ERR;
	}
			
	y = lcdWriteReg(CONST_ADDR_I2C_DEV_ADDR, addr);
	y |= lcdSpecialFunction(CONST_SPL_FUNC_I2C);
	y |= lcdSpecialFunction(CONST_SPL_FUNC_CPU_RESET);
	
	return y;
}




//Initialize communications, i2c/uart/spi
U8 init_coms(void)
{  U8 x = CONST_ERR;
	
	#if defined(I2C_MODE)
	x = sys_i2c_init();
		
	#elif defined(UART_MODE)
	x = sys_uart_init();
	
	#else//SPI_MODE
	x = sys_spi_init();
	#endif
	
	sys_delay_ms(2000);
    
    return x;
}


U8 main()
{	U8 buf[CONST_LAST_REG_ADDR+1];
	U8 regs[CONST_LAST_REG_ADDR+1];
    U8 x = 0, y;
	
#if defined(I2C_MODE)
		#warning "I2C mode"
#elif defined(UART_MODE)
	#warning "Uart Mode"
#elif defined(SPI_MODE)
	#warning "SPI Mode"
#else
	#error "Error: communications modes undefined, i2c/uart/spi"
	return CONST_ERR;
#endif

    //Initialize the communications, i2c/uart/spi
	x = init_coms();
    
    #if 0
    //Test Backlight
	for(y = 0; y <= CONST_MAX_LCD_BACKLIGHT; y+=10)
	{	x |= lcdBacklight(y);
		
		sprintf(buf, "Backlight = %d/%d", y,CONST_MAX_LCD_BACKLIGHT);
		x |= lcdRowWrite(4,buf);
        sys_delay_ms(500);
	}
	#endif	

    #if 0
    //Test Contrast
	for(y = 0; y <= CONST_MAX_LCD_CONTRAST; y+=10)
	{	x |= lcdContrast(y);
		
		sprintf(buf, "Contrast = %d/%d", y,CONST_MAX_LCD_CONTRAST);
		x |= lcdRowWrite(4,buf);
		sys_delay_ms(500);
	}
	//Reset LCD to get back to the original setting
	x |= lcdReset();
    #endif
	
    

    #if 0
    //Example that writes to each row
    //Erase
    for(y = 1; y <= CONST_LCD_NUM_ROWS; y++)
    {	sprintf(buf, " ");
        x |= lcdRowWrite(y,buf);
    }
   
	//Write
	for(y = 1; y <= CONST_LCD_NUM_ROWS; y++)
	{	sprintf(buf, "Row %d",y);
		x |= lcdRowWrite(y,buf);
	}
     #endif
    
    
    #if 1
    //Example that reads PCB and Firmware Versions
    x = lcdBurstReadReg(0, 5, regs);
    if(x == CONST_SUCCESS)
    {   //Erase all the rows
        for(y = 1; y <= CONST_LCD_NUM_ROWS; y++)
        {	sprintf(buf, " ");
            x |= lcdRowWrite(y,buf);
        }
    
        //check ID
        if(regs[CONST_ADDR_ID_REV] == CONST_ADDR_ID_REG_DEF)
        {   sprintf(buf, "PCB Rev %d.%d",regs[CONST_ADDR_PCB_REV]>>4,regs[CONST_ADDR_PCB_REV]&0x0f);
            x |= lcdRowWrite(1,buf);
            sprintf(buf, "FW Rev %d.%d",regs[CONST_ADDR_FW_REV]>>4,regs[CONST_ADDR_FW_REV]&0x0f);
            x |= lcdRowWrite(2,buf);
        }
        else
        {   sprintf(buf, "ID Err %x/%x",regs[CONST_ADDR_ID_REV],CONST_ADDR_ID_REG_DEF);
            x |= lcdRowWrite(1,buf);
        }
    }
    #endif
	
    #if 0
    //For hardware debugging,
    while(1)
    {   
        x = lcdBurstReadReg(0, 2, regs);
        sys_delay_ms(1);
    }
    #endif
    
    sys_delay_ms(1000);
    for(x=5; x != 0; x--)
    {   sprintf(buf, "Reset in %d", x);
        lcdRowWrite(4,buf);
        sys_delay_ms(1000);
    }
    lcdReset();
    sys_delay_ms(1000);
    asm("reset");
    return 0;
}

