#include <xc.h>
#include "project.h"
#include "system.h"


 
//This sample code was tested on a PIC16LF1559 microcontroller, running
//from its internal 16Mhz oscillator using XC8 compiler.
//
//PIC16LF1559 configuration
// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = SWDTEN    // Watchdog Timer Enable (WDT controlled by the SWDTEN bit in the WDTCON register)
#pragma config PWRTE = ON       // Power-up Timer Enable (PWRT enabled)
#pragma config MCLRE = OFF      // MCLR Pin Function Select (MCLR/VPP pin function is digital input)
#pragma config CP = ON          // Flash Program Memory Code Protection (Program memory code protection is enabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled. SBOREN bit is ignored.)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), 1.9V trip point selected.)
#pragma config LPBOR = ON       // Low-Power Brown Out Reset (Low-Power BOR is enabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)

// #pragma config statements should precede project file includes.
// Use project enums instead of #define for ON and OFF.

#include <xc.h>


static void sys_delay_1ms(void);
 
/////////////////////////////////////////////////
//Operating System Specific Functions
/////////////////////////////////////////////////



#if defined(SPI_MODE)
//SPI mode should match LCD setting
//Max clock frequency is 3Mhz
//Chip select should go active(low) at least 300nS before the clock
//activates.
//
U8 sys_spi_init(void)
{	
	//Setup PIC16LF1559 internal oscillator to 16Mhz
    OSCCON  = 0b11111000;   
    
    
    ANSELCbits.ANSC6 = 0; LATC6 = 1; TRISC6 = 0;   //Chip select output
    ANSELCbits.ANSC7 = 0; TRISC7 = 0;   //MOSI output
    ANSELBbits.ANSB4 = 0; TRISB4 = 1;   //MISO input
    ANSELBbits.ANSB6 = 0; LATB6 = 0; TRISB6 = 0;   //Clock output
	
    #ifdef USE_HW_MODULES
    SSPCON1 = 0x00;
    SSPCON3 = 0x00;
    SSPSTAT = 0b01000000;   //shift on transition from idle to active clock
    SSPCON1 = 0b00100001;   //SPI master, FOSC/16 clock, clock idle low
    #endif
            
    return CONST_SUCCESS;
}

void sys_spi_cs(U8 x)
{   if(x)
    {   SPI_CS_HIGH;
    }
    else
    {   SPI_CS_LOW;
    }
}
//Clock one byte, mode 0
U8 sys_spi_byte(U8 tx, U8 *rx)
{   U8 x,data=0;

#ifdef USE_HW_MODULES
    //Start
    SSPBUF = tx;
    
    //Wait for transfer to complete
    while(!SSPSTATbits.BF)
    {}
    
    *rx = SSPBUF;
    
#else    
    //Most significant bit first
    for(x=0; x < 8; x++)
    {   data <<=1;
    
        if(tx & 0x80)
        {   SPI_MOSI_HIGH;   
        }
        else
        {   SPI_MOSI_LOW;
        }
        
        if(IS_SPI_MISO_HIGH)
        {   data |= 0x01;
        }

        SPI_DELAY;
        
        SPI_CLK_HIGH;
        SPI_DELAY;
        SPI_CLK_LOW;
        SPI_DELAY;
        
        tx<<=1;
    }
	*rx = data;
#endif
    return CONST_SUCCESS;
} 

#endif




#if defined(UART_MODE)
//Setup the PIC16LF1559 hardware UART for 8 data bits, one stop bit, no flow control
U8 sys_uart_init(void)
{	
	//Setup PIC16LF1559 internal oscillator to 16Mhz
    OSCCON  = 0b11111000;   

	//Rx pin input
    ANSELBbits.ANSB5 = 0;  TRISB5 = 1;
    
    //Tx pin output
    ANSELBbits.ANSB7 = 0; TRISB7 = 0;
    
    
    #ifdef USE_HW_MODULES
    TXSTA = 0x00;
    RCSTA = 0x00;
    SPBRGL = 103;   //9600 baud
    
	BAUDCON = 0b0000000; //BRG16(bit3) is cleared
    TXSTA = 0b10100100; //BRGH(bit2) is set
    RCSTA = 0b10010000;
#else
#error "Bit banging uart not implemented in this example, too dependent on compiler"
#endif
    
    return CONST_SUCCESS;
}

//Transmit 1 byte out of the uart
//Return CONST_SUCCESS if everything is ok
U8 sys_uart_tx_byte(U8 data)
{	U8 timeout = 0;
    
    //Wait for data to go out before writing next byte
    while(!TXSTAbits.TRMT)
    {   if(++timeout == 0xff)
        {   return CONST_ERR;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }

    TXREG = data;
    
    return CONST_SUCCESS;
}


U8 sys_uart_rx_byte(U8 *data)
{	U8 timeout = 0;

    //Wait for data
    while(!PIR1bits.RCIF)
    {   if(++timeout == 0xff)
        {   return CONST_ERR;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
        
    *data = RCREG;
	return CONST_SUCCESS;
}
 
#endif



#if defined(I2C_MODE)

static void sys_i2c_start(void);
static void sys_i2c_stop(void);
static U8 sys_i2c_write_byte(U8 data);
static U8 sys_i2c_read_byte(U8 *data, U8 ack);
static U8 sys_i2c_wait_clock_stretch(void);

//Configure i2c for a baud rate of up to 400Khz
//Setup at least 5mS of clock stretching
//Setup the slave address to match the LCD
//
//Note: If I2C master doesn't support clock stretching, just modify sys_i2c_write and sys_i2c_read
//to send one byte at a time and pause for 5mS between bytes. The LCD usually holds the clock
//for no more than 100uS, so 5mS should be plenty
//
//Don't forget that SDA and SCL required pull up resistors to 3.3V, 2K is a good choice
//In this example the internal pull ups are enabled. 
U8 sys_i2c_init(void)
{   
    //Setup PIC16LF1559 internal oscillator to 16Mhz
    OSCCON  = 0b11111000;   
    
    //Global pullup enable
    OPTION_REGbits.nWPUEN = 0;
    
    //SDA pin a digital input, pull up enabled, latch set low
    ANSB4=0; TRISB4=1; LATB4=0; WPUB4=1;
    
    //SCL pin a digital input, pull up enabled, latch set low
    ANSB6=0; TRISB6=1; LATB6=0; WPUB6=1;
    

#ifdef USE_HW_MODULES
    SSPSTAT = 0x00; 
    SSPADD = 39;    //100Khz baud 
    SSPCON1 = 0b00101000;   //I2C master mode
    SSPCON2 = 0b00010000;   //ACKEN set
    SSPCON3 = 0b00000000;
#endif
    return CONST_SUCCESS;
}



static void sys_i2c_start(void)
{   U8 timeout;   
    
#ifdef USE_HW_MODULES
    I2C_CLEAR_INT;
    SSPCON2bits.SEN = 1;
    
    //Wait for start to complete
    timeout = 0;
    while(!IS_I2C_INT)
    {   if(++timeout == 250)
        {   break;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
#else
    I2C_SET_SCL
    I2C_SET_SDA
    I2C_DELAY
    I2C_CLR_SDA
    I2C_DELAY
    I2C_CLR_SCL
    I2C_DELAY
#endif
}

static void sys_i2c_stop(void)
{   U8 timeout;
    
#ifdef USE_HW_MODULES
    I2C_CLEAR_INT;
    SSPCON2bits.PEN = 1;
    
    //Wait for stop to complete
    timeout = 0;
    while(!IS_I2C_INT)
    {   if(++timeout == 250)
        {   break;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
#else
    I2C_CLR_SDA
    I2C_DELAY
    I2C_SET_SCL
    I2C_DELAY
    
    //Wait for clock stretching, SCL will stay low
    //while the slave is clock stretching
    sys_i2c_wait_clock_stretch();

    I2C_SET_SDA
    I2C_DELAY
#endif
}



//Wait for slave to release the clock after a write or a read
static U8 sys_i2c_wait_clock_stretch(void)
{   
#ifdef USE_BIT_BANGING
    
    U8 timeout = 0;

    while(!IS_SCL_PIN_HIGH)
    {   I2C_DELAY;
        
        if(++timeout == 0xff)
        {   return CONST_ERR;
        }
    }
#endif
    return CONST_SUCCESS;
}

static U8 sys_i2c_write_byte(U8 data)
{   U8 i, ack, timeout;

#ifdef USE_HW_MODULES
    I2C_CLEAR_INT;
    SSPCON2bits.RCEN = 0;
    SSPBUF = data;
    
    
    //Wait for done
    timeout = 0;
    while(!IS_I2C_INT)
    {   if(++timeout == 250)
        {   CONST_ERR;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
    ack = SSPCON2bits.ACKSTAT;
#else
    for( i = 0; i < 8; i++ )
    {   if(data & 0x80)   //most-significant bit first
        {   I2C_SET_SDA;
        }
        else
        {   I2C_CLR_SDA;
        }
    
        I2C_DELAY
        I2C_SET_SCL
        
        //Wait for clock stretching, SCL will stay low
        //while the slave is clock stretching
        if(sys_i2c_wait_clock_stretch() != CONST_SUCCESS)
        {   return CONST_ERR;
        }
        
        I2C_DELAY
        I2C_CLR_SCL
        data <<= 1;
    }
    
    //Read the ACK/NACK
    I2C_SET_SDA
    I2C_DELAY
    I2C_SET_SCL
    I2C_DELAY
    
    ack = IS_SDA_PIN_HIGH;
    I2C_CLR_SCL
#endif
    if(ack)
    {   return CONST_ERR;
    }
    else
    {   return CONST_SUCCESS;
    }
}



static U8 sys_i2c_read_byte(U8 *data, U8 ack)
{   U8 i, x = 0, timeout;

#ifdef USE_HW_MODULES
    I2C_CLEAR_INT;
    SSPCON2bits.RCEN = 1;
    
    //Wait for done
    timeout = 0;
    while(!IS_I2C_INT)
    {   if(++timeout == 250)
        {   CONST_ERR;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
    *data = SSPBUF;
    I2C_CLEAR_INT;
    
    //Start ACK/NACK
    if(ack)
    {   SSPCON2bits.ACKDT = 0;
    }
    else
    {   SSPCON2bits.ACKDT = 1;
    }
    SSPCON2bits.ACKEN = 1;
    
    
    //Wait for done
    timeout = 0;
    while(!IS_I2C_INT)
    {   if(++timeout == 250)
        {   CONST_ERR;
        }
        I2C_DELAY;I2C_DELAY;I2C_DELAY;I2C_DELAY;
    }
    
#else
    for( i = 0; i < 8; i++ )
    {   x <<= 1;
    
        I2C_SET_SDA
        I2C_DELAY
        I2C_SET_SCL
        
        //Wait for clock stretching, SCL will stay low
        //while the slave is clock stretching
        if(sys_i2c_wait_clock_stretch() != CONST_SUCCESS)
        {   return CONST_ERR;
        }
        I2C_DELAY
        
        if(IS_SDA_PIN_HIGH) 
        {   x |= 0x01;
        }
        
        I2C_DELAY;
        I2C_CLR_SCL
    }
    
    
    //ACK/NACK
    if(ack)
    {   I2C_CLR_SDA
    }
    else
    {   I2C_SET_SDA
    }
    I2C_DELAY
    I2C_SET_SCL
    I2C_DELAY
    I2C_CLR_SCL
            
    *data = x;
#endif
    return CONST_SUCCESS;
}

U8 sys_i2c_write(U8 dev_addr, U8 *buf, U8 len)
{	U8 x = 0, ok=CONST_SUCCESS;

    sys_i2c_start();
    ok |= sys_i2c_write_byte(dev_addr<<1);
    while(len && (ok==CONST_SUCCESS))
    {   ok |= sys_i2c_write_byte(buf[x]);
        len--;
        x++;
    }
    sys_i2c_stop();
    return ok;
}


U8 sys_i2c_read(U8 dev_addr, U8 *buf, U8 len)
{	U8 x = 0, ack=1, ok=CONST_SUCCESS;

    sys_i2c_start();
    ok |= sys_i2c_write_byte(dev_addr<<1 | 0x01);
    while(len && (ok==CONST_SUCCESS))
    {   //NACK on last byte
        if(len == 1)
        {   ack = 0;
        }
        ok |= sys_i2c_read_byte(buf+x,ack);
       
        len--;
        x++;
    }
    sys_i2c_stop();
    return ok;
}

#endif




void sys_delay_ms(U16 x)
{   while(x)
    {   sys_delay_1ms();
        x--;
    }
}

//LCD usually clock stretches for less than 200uS, 1mS should be plenty
static void sys_delay_1ms(void)
{   U8 x;

    for(x = 0; x < 250; x++)
    {   I2C_DELAY
    }
}

#if 0
void sys_i2c_dbg_io(void)
{   sys_i2c_init();
  

    while(1)
    {   I2C_CLR_SCL
        I2C_DELAY
        
        I2C_CLR_SDA
        I2C_DELAY
            
        I2C_SET_SCL
        I2C_DELAY
        
        I2C_SET_SDA
        I2C_DELAY
    }
}   
#endif
